from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.programming.regular_expressions import GeneralRegularExpression

import re


# TODO: Create a CaseTransformation or similar
def snake_case_to_upper_camel_case(snake_str: str):
    """
    Turn the provided 'snake_str' text (that should be like
    'this_is_the_variable_name') into a upper camel case
    name (that would be, for the previous example, 
    'ThisIsTheVariableName').
    """
    if not PythonValidator.is_string(snake_str) or not GeneralRegularExpression.SNAKE_CASE.parse(snake_str):
        raise Exception('The provided "snake_str" parameter is not a valid string.')
    
    return ''.join(word.title() for word in snake_str.split('_'))
    
def snake_case_to_lower_camel_case(snake_str: str):
    """
    Turn the provided 'snake_str' text (that should be like
    'this_is_the_variable_name') into a lower camel case
    name (that would be, for the previous example, 
    'thisIsTheVariableName').
    """
    if not PythonValidator.is_string(snake_str) or not GeneralRegularExpression.SNAKE_CASE.parse(snake_str):
        raise Exception('The provided "snake_str" parameter is not a valid string.')

    parts = snake_str.split('_')

    return parts[0] + ''.join(word.capitalize() for word in parts[1:])

def upper_camel_case_to_snake_case(upper_camel_case_str: str):
    """
    Turn the provided 'upper_camel_case_str' string (that
    should be like 'ThisVariable') into a snake case string
    (that would be, for the previous example, 'this_variable).
    """
    if not PythonValidator.is_string(upper_camel_case_str) or not GeneralRegularExpression.UPPER_CAMEL_CASE.parse(upper_camel_case_str):
        raise Exception('The provided "upper_camel_case_str" is not a valid string.')

    return re.sub(r'([a-z])([A-Z])', r'\1_\2', upper_camel_case_str).lower()

def lower_camel_case_to_snake_case(lower_camel_case_str: str):
    """
    Turn the provided 'lower_camel_case_str' string (that
    should be like 'thisVariable') into a snake case string
    (that would be, for the previous example, 'this_variable).
    """
    if not PythonValidator.is_string(lower_camel_case_str) or not GeneralRegularExpression.LOWER_CAMEL_CASE.parse(lower_camel_case_str):
        raise Exception('The provided "upper_camel_case_str" is not a valid string.')
    
    return re.sub(r'([a-z])([A-Z])', r'\1_\2', lower_camel_case_str).lower()

# TODO: Implement more conversions


from yta_general_utils.programming.validator import PythonValidator
from typing import Union


class ErrorMessage:
    """
    Class to encapsulate the different error
    messages we need.
    """

    @staticmethod
    def parameter_is_not_a_class(
        parameter_name: str
    ) -> str:
        return f'The provided "{parameter_name}" parameter is not a class.'
    
    @staticmethod
    def parameter_not_provided(
        parameter_name: str
    ) -> str:
        return f'The parameter "{parameter_name}" was not provided.'
    
    @staticmethod
    def parameter_is_not_string(
        parameter_name: str
    ) -> str:
        return f'The parameter "{parameter_name}" provided is not a string.'
    
    @staticmethod
    def parameter_is_not_boolean(
        parameter_name: str
    ) -> str:
        return f'The "{parameter_name}" parameter is not boolean.'
    
    @staticmethod
    def parameter_is_not_int(
        parameter_name: str
    ) -> str:
        return f'The "{parameter_name}" parameter is not an int.'
    
    @staticmethod
    def parameter_is_not_positive_int(
        parameter_name: str
    ) -> str:
        return f'The "{parameter_name}" parameter is not a positive int.'

    @staticmethod
    def parameter_is_not_float(
        parameter_name: str
    ) -> str:
        return f'The "{parameter_name}" parameter is not a float.'
    
    @staticmethod
    def parameter_is_not_positive_float(
        parameter_name: str
    ) -> str:
        return f'The "{parameter_name}" parameter is not a positive float.'

    @staticmethod
    def parameter_is_not_dict(
        parameter_name: str
    ) -> str:
        return f'The "{parameter_name}" parameter is not a dict.'
    
    @staticmethod
    def parameter_is_not_callable(
        parameter_name: str
    ) -> str:
        return f'The "{parameter_name}" parameter is not callable.'

    @staticmethod
    def parameter_is_not_positive_number(
        parameter_name: str
    ) -> str:
        return f'The parameter "{parameter_name}" provided is not a valid and positive number.'
    
    @staticmethod
    def parameter_is_file_that_doesnt_exist(
        parameter_name: str
    ) -> str:
        return f'The "{parameter_name}" parameter provided is not a file that exists.'
    
    @classmethod
    def parameter_is_not_file_of_file_type(
        parameter_name: str,
        file_type: 'FileType'
    ) -> str:
        return f'The "{parameter_name}" provided is not a {file_type.value} filename.'
    
    @staticmethod
    def parameter_is_not_valid_url(
        parameter_name: str
    ) -> str:
        return f'The provided "{parameter_name}" parameter is not a valid url.'

    @staticmethod
    def parameter_is_not_list_of_string(
        parameter_name: str
    ) -> str:
        return f'The provided "{parameter_name}" parameter is not a list of strings.'

    @staticmethod
    def parameter_is_not_list_of_numbers(
        parameter_name: str
    ) -> str:
        return f'The provided "{parameter_name}" parmaeter is not a list of numbers.'
    
    @staticmethod
    def parameter_is_not_list_of_positive_numbers(
        parameter_name: str
    ) -> str:
        return f'The provided "{parameter_name}" parameter is not a ist of positive numbers.'

    @staticmethod
    def parameter_is_not_list_of_classes(
        parameter_name: str,
    ) -> str:
        return f'The provided "{parameter_name}" parameter is not a list of classes.'

    @staticmethod
    def parameter_is_not_list_of_these_classes(
        parameter_name: str,
        cls: Union[list[Union[type, str]], str, type]
    ) -> str:
        return f'The provided "{parameter_name}" parameter is not a list of classes of this group: {", ".join(_cls_parameter_to_string_classes_array(cls))}'

    @staticmethod
    def parameter_is_not_list_of_instances(
        parameter_name: str
    ) -> str:
        return f'The provided "{parameter_name}" parameter is not a list of instances.'
    
    @staticmethod
    def parameter_is_not_list_of_these_instances(
        parameter_name: str,
        cls: Union[list[Union[type, str]], str, type]
    ) -> str:
        return f'The provided "{parameter_name}" parameter is not a list of instances of this group: {", ".join(_cls_parameter_to_string_classes_array(cls))}'

    @staticmethod
    def parameter_is_not_class_of(
        parameter_name: str,
        cls: Union[list[Union[type, str]], str, type]
    ) -> str:
        return f'The provided "{parameter_name}" parameter is not one of these classes: {", ".join(_cls_parameter_to_string_classes_array(cls))}.'
    
    @staticmethod
    def parameter_is_not_instance_of(
        parameter_name: str,
        cls: Union[list[Union[type, str]], str, type]
    ) -> str:
        """
        Message that indicates that the 'parameter_name'
        is not any of the provided 'classes'.
        """
        return f'The provided "{parameter_name}" parameter is not an instance of any of these classes: {", ".join(_cls_parameter_to_string_classes_array(cls))}.'

    @staticmethod
    def keys_are_missing_in_dict(
        name: str,
        missing_keys: list[str]
    ) -> str:
        """
        Message that indicates that the dict parameter
        with the given 'name' doesn't have the also given
        'missing_keys' keys set.
        """
        return f'These keys are missing in the "{name}" dict: {", ".join(missing_keys)}'

    @staticmethod
    def parameter_is_not_name_of_ytaenum_class(
        name: str,
        enum
    ) -> str:
        return f'The provided YTAEnum name "{name}" is not a valid {enum.__class__.__name__} YTAEnum name.'
    
    @staticmethod
    def parameter_is_not_value_of_ytaenum_class(
        value: any,
        enum
    ) -> str:
        return f'The provided YTAEnum value "{value}" is not a valid {enum.__class__.__name__} YTAEnum value.'
    
    @staticmethod
    def parameter_is_not_name_nor_value_of_ytaenum_class(
        name_or_value: any,
        enum
    ) -> str:
        return f'The provided YTAEnum name or value "{name_or_value}" is not a valid {enum.__class__.__name__} YTAEnum name or value.'
    
    @staticmethod
    def parameter_is_not_name_nor_value_nor_enum_of_ytaenum_class(
        name_or_value_or_enum: any,
        enum
    ) -> str:
        return f'The provided YTAEnum name, value or instance "{name_or_value_or_enum}" is not a valid {enum.__class__.__name__} YTAEnum name, value or instance.'
    
def _cls_parameter_to_string_classes_array(
    cls: Union[list[Union[type, str]], str, type]
) -> list[str]:
    """
    Transform the given 'cls' class or class array
    to an array of string class name(s).
    """
    # To list
    cls = (
        [cls]
        if not PythonValidator.is_list(cls) else
        cls
    )

    # To str array
    return [
        (
            cls_item
            if PythonValidator.is_string(cls_item) else
            cls_item.__name__ # the name of the class
        ) for cls_item in cls
    ]


"""
When we handle files with our system we obtain them
in different formats. Sometimes we get them from urls
so they are just a bytes array, and sometimes we 
obtain an image, for example, that has been previously
loaded with its corresponding library.

We try to treat all those files in the same way so we
have created this class to interact with them and make
easier the way we handle them.
"""
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.file.reader import FileReader
from dataclasses import dataclass
from typing import Union

import io


@dataclass
class FileReturn:
    """
    This dataclass has been created to handle a file
    that has been created or downloaded, so we are
    able to return the file itself and also the 
    filename in the same return.
    """

    # TODO: Set valid types
    file: Union[bytes, bytearray, io.BytesIO, any]
    """
    The file content as raw as it was obtained by
    our system, that could be binary or maybe an
    actually parsed file.
    """
    type: Union[FileTypeX, None]
    """
    The type of the obtained file.
    """
    filename: str
    """
    The filename of the obtained file.
    """

    @property
    def file_type(self) -> Union[FileTypeX, None]:
        """
        Get the FileTypeX associated to the
        'output_filename' extension if existing and
        valid.
        """
        return FileTypeX.get_type_from_filename(self.filename)

    @property
    def file_converted(self):
        """
        The file parsed according to its type. This
        can be the same as 'file' attribute if it
        was obtained in a converted format.
        """
        # Sometimes the file that has been set is
        # already converted, so we just send it
        # as it is
        if not PythonValidator.is_instance(self.file, [bytes, bytearray, io.BytesIO]):
            return self.file
        
        if self.type is None:
            # TODO: What about this (?)
            import warnings
            warnings.warn('The type is None so we do not actually know the file type. Returning it raw.')
            return self.file

        return FileReader.parse_file_content(self.file, self.type)

    def __init__(
        self,
        file: Union[bytes, bytearray, io.BytesIO, any],
        type: Union[FileTypeX, None],
        filename: str
    ):
        if type is not None:
            type = FileTypeX.to_enum(type)

        self.file = file
        self.type = type
        self.filename = filename




from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.programming.validator.number import NumberValidator
from yta_general_utils.random import Random
from yta_general_utils.math.value_normalizer import ValueNormalizer
from typing import Union

import numpy as np


COORDINATE_MAX_POSITION = (1920 * 4, 1080 * 4)
"""
The maximum accepted position for a coordinate.
This position will be used for validation and
for normalization.
"""
COORDINATE_MIN_POSITION = (-COORDINATE_MAX_POSITION[0], -COORDINATE_MAX_POSITION[1])
"""
The minimum accepted position for a coordinate.
This position will be used for validation and
for normalization.
"""



# TODO: This is the new Coordinate class we want to keep
# when we know it is working properly
class Coordinate:
    """
    Class to represent a point, a position, in a graphic
    defined by both X and Y axis. This coordinate exist
    in our defined virtual limits and it is useful to
    position images or videos within a scene.

    This class is able to normalize and denormalize its
    values according to the scenario.
    """

    position: tuple[float, float] = None
    """
    The coordinate position not normalized. The X and
    Y axis values are within our specific limits 
    defined in this file.
    """
    _normalized_position: tuple[float, float] = None
    """
    The coordinate position but normalized, meaning
    that the X and Y axis are values within the
    [0, 1] range.
    """

    @property
    def x(self):
        return self.position[0]
    
    @property
    def y(self):
        return self.position[1]
    
    @property
    def normalized_position(self):
        """
        The coordinate position but normalized, meaning
        that the X and Y axis are values within the
        [0, 1] range.
        """
        if self._normalized_position is None:
            value_normalizer = ValueNormalizer(COORDINATE_MIN_POSITION, COORDINATE_MAX_POSITION)

            self._normalized_position = (
                value_normalizer.normalize(self.x),
                value_normalizer.normalize(self.y)
            )

        return self._normalized_position
    
    @property
    def x_normalized(self):
        return self.normalized_position[0]
    
    @property
    def y_normalized(self):
        return self.normalized_position[1]
    
    @property
    def as_tuple(self):
        """
        Return the coordinate as a not normalized tuple
        ('x', 'y').
        """
        return (self.x, self.y)
    
    @property
    def as_tuple_normalized(self):
        """
        Return the coordinate as a normalized tuple
        ('x', 'y').
        """
        return (self.x_normalized, self.y_normalized)
    
    @property
    def as_array(self):
        """
        Return the coordinate as a not normalized array
        ['x', 'y'].
        """
        return [self.x, self.y]
    
    @property    
    def as_array_normalized(self):
        """
        Return the coordinate as a normalized array
        ['x', 'y'].
        """
        return [self.x_normalized, self.y_normalized]
    
    def __init__(self, position: Union[tuple[int, int], list[int, int]], is_normalized: bool = False):
        validate_coordinate_position(position, is_normalized)

        value_normalizer = ValueNormalizer(COORDINATE_MIN_POSITION, COORDINATE_MAX_POSITION)

        position = (
            value_normalizer.denormalize(position[0]),
            value_normalizer.denormalize(position[1])
        ) if is_normalized else (
            position[0],
            position[1]
        )
        
        self.position = position
    
    @staticmethod
    def generate(n: int = 1, lower_limit: Union[tuple[int, int], list[int, int]] = (0, 0), upper_limit: Union[tuple[int, int], list[int, int]] = (1920, 1080)):
        """
        Generate 'n' coordinates with random values
        between the 'lower_limit' and the 'upper_limit' for
        that are returned as an array of instances.

        The 'n', which is the amount of coordinates to create,
        is limited to the interval [1, 100].
        """
        if not NumberValidator.is_number_between(n, 1, 100):
            raise Exception(f'The provided "amount" parameter "{str(n)}" is not a number between 1 and 100.')
        
        # This is not actually a coordinate position but
        # a coordinate limit, so it works similar
        validate_coordinate_position(lower_limit)
        validate_coordinate_position(upper_limit)
        
        return [
            Coordinate((
                Random.int_between(lower_limit[0], upper_limit[0]),
                Random.int_between(lower_limit[1], upper_limit[1])
            ))
            for _ in range(n)
        ]

# TODO: Remove this Coordinate when the one above
# has been refactored and it is working perfectly

NORMALIZATION_MIN_VALUE = -10_000
"""
The lower limit for the normalization process.
"""
NORMALIZATION_MAX_VALUE = 10_000
"""
The upper limit for the normalization process.
"""

# TODO: Is Coordinate class actually needed (?)
# class Coordinate:
#     """
#     Class to represent a coordinate point ('x', 'y').
#     """
#     position: tuple = None
#     """
#     The ('x', 'y') tuple containing the position
#     coordinate.
#     """
#     _is_normalized: bool = False
#     """
#     Internal function to know if it has been normalized
#     or not.
#     """

#     @property
#     def x(self):
#         return self.position[0]
    
#     @property
#     def y(self):
#         return self.position[1]
    
#     @property
#     def is_normalized(self):
#         return self._is_normalized

#     def __init__(self, x: float, y: float, is_normalized: bool = False):
#         if not NumberValidator.is_number_between(x, NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE) or not NumberValidator.is_number_between(y, NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE):
#             raise Exception(f'The "x" and "y" parameters must be values between {str(NORMALIZATION_MIN_VALUE)} and {str(NORMALIZATION_MAX_VALUE)} and "{str(x)}, {str(y)}" provided.')
        
#         if not PythonValidator.is_boolean(is_normalized):
#             raise Exception('The "is_normalized" parameter must be a boolean value.')
        
#         self.position = (x, y)
#         self._is_normalized = is_normalized

#     def get_x(self):
#         """
#         Return the 'x' value.
#         """
#         return self.x
    
#     def get_y(self):
#         """
#         Return the 'y' value.
#         """
#         return self.y

#     def as_tuple(self):
#         """
#         Return the coordinate as a tuple ('x', 'y').
#         """
#         return Coordinate.to_tuple(self)
    
#     def as_array(self):
#         """
#         Return the coordinate as an array ['x', 'y'].
#         """
#         return Coordinate.to_array(self)

#     def normalize(self):
#         """
#         Normalize the coordinate by turning the values into
#         a range between [0.0, 1.0]. This will be done if the
#         values have not been normalized previously.
#         """
#         if not self._is_normalized:
#             self.position = Coordinate.normalize_tuple(self.position)
#             self._is_normalized = True

#         return self

#     def denormalize(self):
#         """
#         Denormalize the coordinate values by turning them
#         from normalized values to the real ones. This will
#         be done if the values have been normalized 
#         previously.
#         """
#         if self._is_normalized:
#             self.position = Coordinate.denormalize_tuple(self.position)
#             self._is_normalized = False

#         return self

#     @staticmethod
#     def to_tuple(coordinate):
#         """
#         Turn the provided 'coordinate' to a tuple like ('x', 'y').
#         """
#         return coordinate.position
    
#     @staticmethod
#     def to_array(coordinate):
#         """
#         Turn the provided 'coordinate' to an array like ['x', 'y'].
#         """
#         return [coordinate.x, coordinate.y]

#     @staticmethod
#     def generate(amount: int = 1):
#         """
#         Generate 'amount' coordinates with random values
#         between [0, 1920] for the 'x' and [0, 1080] for
#         the 'y', that are returned as an array of instances.

#         The 'amount' parameter is limited to the interval 
#         [1, 100].
#         """
#         if not NumberValidator.is_number_between(amount, 1, 100):
#             raise Exception(f'The provided "amount" parameter "{str(amount)}" is not a number between 1 and 100.')
        
#         return Coordinate(random_int_between(0, 1920), random_int_between(0, 1080))
    
#     @staticmethod
#     def to_numpy(coordinates: list['Coordinate']):
#         """
#         Convert a list of Coordinates 'coordinates' to
#         numpy array to be able to work with them.

#         This method does the next operation:
#         np.array([[coord.x, coord.y] for coord in coordinates])
#         """
#         if not PythonValidator.is_list(coordinates):
#             if not PythonValidator.is_instance(coordinates, Coordinate):
#                 raise Exception('The provided "coordinates" parameter is not a list of NormalizedCoordinates nor a single NormalizedCoordinate instance.')
#             else:
#                 coordinates = [coordinates]
#         elif any(not PythonValidator.is_instance(coordinate, Coordinate) for coordinate in coordinates):
#             raise Exception('At least one of the provided "coordinates" is not a NormalizedCoordinate instance.')

#         return np.array([coordinate.as_array() for coordinate in coordinates])
    
#     @staticmethod
#     def normalize_tuple(coordinate: tuple):
#         """
#         Normalize the provided 'coordinate' by applying
#         our normalization limits. This means turning the
#         non-normalized 'coordinate' to a normalized one
#         (values between 0.0 and 1.0).
#         """
#         return (
#             Math.normalize(coordinate[0], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE),
#             Math.normalize(coordinate[1], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE)
#         )
    
#     @staticmethod
#     def denormalize_tuple(coordinate: tuple):
#         """
#         Denormalize the provided 'coordinate' by applying
#         our normalization limits. This means turning the 
#         normalized 'coordinate' (values between 0.0 and
#         1.0) to the not-normalized ones according to our
#         normalization limits.
#         """
#         return (
#             Math.denormalize(coordinate[0], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE),
#             Math.denormalize(coordinate[1], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE)
#         )
    
#     @staticmethod
#     def is_valid(coordinate: tuple):
#         """
#         Check if the provided 'coordinate' is valid or not.
#         A valid coordinate is a tuple with two elements that
#         are values between our normalization limits.
#         """
#         if not PythonValidator.is_instance(coordinate, 'Coordinate') and (not PythonValidator.is_tuple(coordinate) or len(coordinate) != 2 or not NumberValidator.is_number_between(coordinate[0], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE) or not NumberValidator.is_number_between(coordinate[1], NORMALIZATION_MIN_VALUE, NORMALIZATION_MAX_VALUE)):
#             return False
        
#         return True

#     @staticmethod
#     def validate(coordinate: tuple, parameter_name: str):
#         """
#         Validate if the provided 'coordinate' is a coordinate
#         with values between our normalization limits.
#         """
#         if not Coordinate.is_valid(coordinate):
#             raise Exception(f'The provided "{parameter_name}" parameter is not a valid tuple of 2 elements that are values between our limits {str(NORMALIZATION_MIN_VALUE)} and {str(NORMALIZATION_MAX_VALUE)}. Please, provide a valid coordinate.')


def validate_coordinate_position(position: Union[tuple[int, int], list[int, int]], is_normalized: bool = False):
    """
    Validate if a coordinate position is a tuple or list
    of 2 elements within our valid range.
    """
    if (not PythonValidator.is_tuple(position) and not PythonValidator.is_list(position)) or len(position) != 2:
        raise Exception('The provided "position" parameter is not a tuple nor a list of 2 elements.')
    
    if is_normalized:
        if not NumberValidator.is_number_between(position[0], 0, 1) or not NumberValidator.is_number_between(position[1], 0, 1):
            raise Exception(f'The provided "position" ({position[0], position[1]}) is out of the normalized range, which is from (0, 0) to (1, 1)')
    else:
        if not NumberValidator.is_number_between(position[0], COORDINATE_MIN_POSITION[0], COORDINATE_MAX_POSITION[0]) or not NumberValidator.is_number_between(position[1], COORDINATE_MIN_POSITION[1], COORDINATE_MAX_POSITION[1]):
            raise Exception(f'The provided "position" ({position[0], position[1]}) is out of range, which is from ({COORDINATE_MIN_POSITION[0]}, {COORDINATE_MIN_POSITION[1]}) to ({COORDINATE_MAX_POSITION[0]}, {COORDINATE_MAX_POSITION[1]})')

def coordinates_to_numpy(coordinates: list['Coordinate']):
    """
    Convert a list of Coordinates 'coordinates' to
    numpy array to be able to work with them.

    This method does the next operation:
    np.array([[coord.x, coord.y] for coord in coordinates])
    """
    if not PythonValidator.is_list(coordinates):
        if not PythonValidator.is_instance(coordinates, Coordinate):
            raise Exception('The provided "coordinates" parameter is not a list of NormalizedCoordinates nor a single NormalizedCoordinate instance.')
        else:
            coordinates = [coordinates]
    elif any(not PythonValidator.is_instance(coordinate, Coordinate) for coordinate in coordinates):
        raise Exception('At least one of the provided "coordinates" is not a NormalizedCoordinate instance.')

    return np.array([coordinate.as_array() for coordinate in coordinates])


from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.temp import Temp
from yta_general_utils.random import Random
from yta_general_utils.programming.validator import PythonValidator
from typing import Union

import os


class FileExtension(Enum):
    """
    Enum class to encapsulate all the file
    extensions that we are able to handle.
    """

    # IMAGE
    PNG = 'png'
    """
    Portable Network Graphics
    """
    JPEG = 'jpeg'
    """
    Joint Photographic Experts Group
    """
    JPG = 'jpg'
    """
    Joint Photographic Experts Group
    """
    WEBP = 'webp'
    """
    Web Picture
    """
    BMP = 'bmp'
    """
    Bitmap Image File
    """
    GIF = 'gif'
    """
    Graphics Interchange Format
    """
    TIFF = 'tiff'
    """
    Tagged Image File
    """
    PSD = 'psd'
    """
    Photoshop Document
    """
    PDF = 'pdf'
    """
    Portable Document Format
    """
    EPS = 'eps'
    """
    Encapsulated Postcript
    """
    AI = 'ai'
    """
    Adobe ILlustrator Document
    """
    INDD = 'indd'
    """
    Adobe Indesign Document
    """
    RAW = 'raw'
    """
    Raw Image Formats
    """
    CDR = 'cdr'
    """
    Corel Draw
    """
    # AUDIO
    WAV = 'wav'
    """
    Waveform Audio
    """
    MP3 = 'mp3'
    """
    MPEG Audio Layer 3.
    """
    M4A = 'm4a'
    """
    MPEG-4 Audio
    """
    FLAC = 'flac'
    """
    Free Lossless Audio Codec.
    """
    WMA = 'wma'
    """
    Windows Media Audio
    """
    AAC = 'aac'
    """
    Advanced Audio Coding
    """
    # VIDEO
    MOV = 'mov'
    """
    Apple video
    """
    MP4 = 'mp4'
    """
    MPEG-4
    """
    WEBM = 'webm'
    """
    Developed by Google, subgroup of the open and standard Matroska Video Container (MKV)
    """
    AVI = 'avi'
    """
    Audio Video Interleave
    """
    WMV = 'wmv'
    """
    Windows Media Video
    """
    AVCHD = 'avchd'
    """
    Advanced Video Coding High Definition
    """
    FVL = 'flv'
    """
    Flash Video
    """
    # SUBTITLES
    SRT = 'srt'
    """
    Srt subtitle file extension.

    This is the format:
    1
    00:00:00,000 --> 00:00:02,500
    Welcome to the Example Subtitle File!

    """
    JSON3 = 'json3'
    """
    Json3 subtitle file extension
    """
    SRV1 = 'srv1'
    """
    Srv1 subtitle file extension
    """
    SRV2 = 'srv2'
    """
    Srv2 subtitle file extension
    """
    SRV3 = 'srv3'
    """
    Srv3 subtitle file extension
    """
    TTML = 'ttml'
    """
    Ttml subtitle file extension
    """
    VTT = 'vtt'
    """
    Vtt subtitle file extension
    """
    # TEXT
    TXT = 'txt'
    """
    Txt text file extension
    """
    CSV = 'csv'
    """
    Csv text file extension
    """
    JSON = 'json'
    """
    Json text file extension
    """
    XML = 'xml'
    """
    Xml text file extension
    """
    HTML = 'html'
    """
    Html text file extension
    """
    MD = 'md'
    """
    Md text file extension
    """
    LOG = 'log'
    """
    Log text file extension
    """
    INI = 'ini'
    """
    Ini text file extension
    """
    YAML = 'yaml'
    """
    Yaml text file extension
    """
    YML = 'yml'
    """
    Yml text file extension
    """


    # TODO: Add some utils  
    @staticmethod
    def is_filename_valid(
        filename: str
    ):
        """
        Check if the provided 'filename' is valid according
        to this file extension.
        """
        return is_filename_valid_for_file_extension(filename, FileExtension)
    
    def get_temp_filename(
        self,
        filename: Union[str, None] = None
    ) -> str:
        """
        Get a temporary random filename with this file
        extension.

        The 'filename', if provided, will be forced to
        have this file extension but keeping the rest.
        If it is not provided, it will be a randomly
        generated string with this file extension.
        """
        return get_temp_filename_for_file_extension(filename, self)

class ImageFileExtension(Enum):
    """
    Enum class to encapsulate all existing image file extensions.
    """
    
    PNG = FileExtension.PNG.value
    """
    Portable Network Graphics
    """
    JPEG = FileExtension.JPEG.value
    """
    Joint Photographic Experts Group
    """
    JPG = FileExtension.JPG.value
    """
    Joint Photographic Experts Group
    """
    WEBP = FileExtension.WEBP.value
    """
    Web Picture
    """
    BMP = FileExtension.BMP.value
    """
    Bitmap Image File
    """
    GIF = FileExtension.GIF.value
    """
    Graphics Interchange Format
    """
    TIFF = FileExtension.TIFF.value
    """
    Tagged Image File
    """
    PSD = FileExtension.PSD.value
    """
    Photoshop Document
    """
    PDF = FileExtension.PDF.value
    """
    Portable Document Format
    """
    EPS = FileExtension.EPS.value
    """
    Encapsulated Postcript
    """
    AI = FileExtension.AI.value
    """
    Adobe Illustrator Document
    """
    INDD = FileExtension.INDD.value
    """
    Adobe Indesign Document
    """
    RAW = FileExtension.RAW.value
    """
    Raw Image Formats
    """
    CDR = FileExtension.CDR.value
    """
    Corel Draw
    """
    # TODO: Add more

    @classmethod
    def default(cls):
        return cls.PNG

    # TODO: Add some utils 
    # TODO: I cannot inherit from FileExtension Enum :( 
    @staticmethod
    def is_filename_valid(
        filename: str
    ):
        """
        Check if the provided 'filename' is valid according
        to this file extension.
        """
        return is_filename_valid_for_file_extension(filename, ImageFileExtension)
    
    def get_temp_filename(
        self,
        filename: Union[str, None] = None
    ) -> str:
        """
        Get a temporary random filename with this file
        extension.

        The 'filename', if provided, will be forced to
        have this file extension but keeping the rest.
        If it is not provided, it will be a randomly
        generated string with this file extension.
        """
        return get_temp_filename_for_file_extension(filename, self)

class AudioFileExtension(Enum):
    """
    Enum class to encapsulate all existing audio file extensions.
    """

    WAV = FileExtension.WAV.value
    """
    Waveform Audio
    """
    MP3 = FileExtension.MP3.value
    """
    MPEG Audio Layer 3.
    """
    M4A = FileExtension.M4A.value
    """
    MPEG-4 Audio
    """
    FLAC = FileExtension.FLAC.value
    """
    Free Lossless Audio Codec.
    """
    WMA = FileExtension.WMA.value
    """
    Windows Media Audio
    """
    AAC = FileExtension.AAC.value
    """
    Advanced Audio Coding
    """
    WEBM = FileExtension.WEBM.value
    # TODO: Add more

    @classmethod
    def default(cls):
        return cls.WAV
    
    # TODO: I cannot inherit from FileExtension Enum :( 
    @staticmethod
    def is_filename_valid(
        filename: str
    ):
        """
        Check if the provided 'filename' is valid according
        to this file extension.
        """
        return is_filename_valid_for_file_extension(filename, AudioFileExtension)
    
    def get_temp_filename(
        self,
        filename: Union[str, None] = None
    ) -> str:
        """
        Get a temporary random filename with this file
        extension.

        The 'filename', if provided, will be forced to
        have this file extension but keeping the rest.
        If it is not provided, it will be a randomly
        generated string with this file extension.
        """
        return get_temp_filename_for_file_extension(filename, self)

class VideoFileExtension(Enum):
    """
    Enum class to encapsulate all existing video file extensions.
    """

    MOV = FileExtension.MOV.value
    """
    Apple video
    """
    MP4 = FileExtension.MP4.value
    """
    MPEG-4
    """
    WEBM = FileExtension.WEBM.value
    """
    Developed by Google, subgroup of the open and standard Matroska Video Container (MKV)
    """
    AVI = FileExtension.AVI.value
    """
    Audio Video Interleave
    """
    WMV = FileExtension.WMV.value
    """
    Windows Media Video
    """
    AVCHD = FileExtension.AVCHD.value
    """
    Advanced Video Coding High Definition
    """
    FVL = FileExtension.FVL.value
    """
    Flash Video
    """
    # TODO: Add more

    @classmethod
    def default(cls):
        return cls.MP4
    
    # TODO: I cannot inherit from FileExtension Enum :( 
    @staticmethod
    def is_filename_valid(
        filename: str
    ):
        """
        Check if the provided 'filename' is valid according
        to this file extension.
        """
        return is_filename_valid_for_file_extension(filename, VideoFileExtension)
    
    def get_temp_filename(
        self,
        filename: Union[str, None] = None
    ) -> str:
        """
        Get a temporary random filename with this file
        extension.

        The 'filename', if provided, will be forced to
        have this file extension but keeping the rest.
        If it is not provided, it will be a randomly
        generated string with this file extension.
        """
        return get_temp_filename_for_file_extension(filename, self)

class SubtitleFileExtension(Enum):
    """
    Enum class to encapsulate all existing subtitle
    file extensions.
    """

    SRT = FileExtension.SRT.value
    """
    Srt subtitle file extension.

    This is the format:
    1
    00:00:00,000 --> 00:00:02,500
    Welcome to the Example Subtitle File!

    """
    JSON3 = FileExtension.JSON3.value
    """
    Json3 subtitle file extension
    """
    SRV1 = FileExtension.SRV1.value
    """
    Srv1 subtitle file extension
    """
    SRV2 = FileExtension.SRV2.value
    """
    Srv2 subtitle file extension
    """
    SRV3 = FileExtension.SRV3.value
    """
    Srv3 subtitle file extension
    """
    TTML = FileExtension.TTML.value
    """
    Ttml subtitle file extension
    """
    VTT = FileExtension.VTT.value
    """
    Vtt subtitle file extension
    """

    @classmethod
    def default(cls):
        return cls.JSON3
    
    # TODO: I cannot inherit from FileExtension Enum :( 
    @staticmethod
    def is_filename_valid(
        filename: str
    ):
        """
        Check if the provided 'filename' is valid according
        to this file extension.
        """
        return is_filename_valid_for_file_extension(filename, SubtitleFileExtension)
    
    def get_temp_filename(
        self,
        filename: Union[str, None] = None
    ) -> str:
        """
        Get a temporary random filename with this file
        extension.

        The 'filename', if provided, will be forced to
        have this file extension but keeping the rest.
        If it is not provided, it will be a randomly
        generated string with this file extension.
        """
        return get_temp_filename_for_file_extension(filename, self)
    
class TextFileExtension(Enum):
    """
    Enum class to encapsulate all existing text
    file extensions.
    """

    TXT = FileExtension.TXT.value
    """
    Txt subtitle file extension
    """
    JSON = FileExtension.JSON.value
    """
    Json text file extension
    """
    XML = FileExtension.XML.value
    """
    Xml text file extension
    """
    HTML = FileExtension.HTML.value
    """
    Html text file extension
    """
    MD = FileExtension.MD.value
    """
    Md text file extension
    """
    LOG = FileExtension.LOG.value
    """
    Log text file extension
    """
    INI = FileExtension.INI.value
    """
    Ini text file extension
    """
    YAML = FileExtension.YAML.value
    """
    Yaml text file extension
    """
    YML = FileExtension.YML.value
    """
    Yml text file extension
    """

    @classmethod
    def default(cls):
        return cls.TXT
    
    # TODO: I cannot inherit from FileExtension Enum :( 
    @staticmethod
    def is_filename_valid(
        filename: str
    ):
        """
        Check if the provided 'filename' is valid according
        to this file extension.
        """
        return is_filename_valid_for_file_extension(filename, TextFileExtension)
    
    def get_temp_filename(
        self,
        filename: Union[str, None] = None
    ) -> str:
        """
        Get a temporary random filename with this file
        extension.

        The 'filename', if provided, will be forced to
        have this file extension but keeping the rest.
        If it is not provided, it will be a randomly
        generated string with this file extension.
        """
        return get_temp_filename_for_file_extension(filename, self)
    
# TODO: I've been able to inherit from another custom
# YTAEnum classes when creating a new one, but here I
# had some troubles recently so I decided to continue
# and fix it later. Please, review it and refactor
# because I'm repeating a lot of code

# These classes above should be used by the ffmpeg_handler and other
# declarations I make in our app to be consistent and reuse the code

class FileType(Enum):
    """
    Enum that represents the different file types and the valid
    extensions we accept for those file types. This Enum is to
    be used when checking filenames parameter.

    For example, we will use this to make sure the filename they
    gave to us is a video file type if we are storing a video 
    file.
    """
    IMAGE = ImageFileExtension.get_all()
    AUDIO = AudioFileExtension.get_all()
    VIDEO = VideoFileExtension.get_all()

class FileTypeX(Enum):
    """
    TODO: Check original FileType
    """

    IMAGE = 'image'
    AUDIO = 'audio'
    VIDEO = 'video'
    SUBTITLE = 'subtitle'
    TEXT = 'text'

    @staticmethod
    def get_type_from_filename(
        filename: str
    ) -> Union['FileTypeX', None]:
        for type in FileTypeX.get_all():
            if type.is_filename_valid(filename):
                return type

        return None

    def is_filename_valid(
        self,
        filename: str
    ):
        """
        Check if the provided 'filename' is valid according
        to this file extension.
        """
        return is_filename_valid_for_file_extension(filename, self.get_file_extension_enum_class())
    
    def get_file_extension_enum_class(
        self
    ) -> Union[VideoFileExtension, ImageFileExtension, AudioFileExtension]:
        """
        Get the file extension YTAEnum class associated
        with this file type YTAEnum instance.
        """
        return {
            FileTypeX.VIDEO: VideoFileExtension,
            FileTypeX.IMAGE: ImageFileExtension,
            FileTypeX.AUDIO: AudioFileExtension,
            FileTypeX.SUBTITLE: SubtitleFileExtension,
            FileTypeX.TEXT: TextFileExtension
        }[self]

    def get_default_file_extension(
        self
    ) -> Union[VideoFileExtension, ImageFileExtension, AudioFileExtension]:
        """
        Get the default file extension of this file type.
        """
        return self.get_file_extension_enum_class().default()

    def get_temp_filename(self, filename: Union[str, None]):
        """
        Get a temporary random filename for this file type.
        """
        return get_temp_filename_for_file_extension(filename, self.get_default_file_extension())

def is_filename_valid_for_file_extension(
    filename: str,
    file_extension_enum_class: Union[FileExtension, TextFileExtension, AudioFileExtension, ImageFileExtension, VideoFileExtension, SubtitleFileExtension] 
):
    """
    Check if the provided 'filename' is valid for
    the also given 'file_extension_enum_class'.
    """
    # TODO: Fix this, 'is_class' is not accepting arrays
    # if not PythonValidator.is_class(file_extension_enum_class, [FileExtension, TextFileExtension, AudioFileExtension, ImageFileExtension, VideoFileExtension, SubtitleFileExtension]):
    #     raise Exception('The "file_extension_enum" is not valid.')

    if filename is None:
        return False
    
    if not PythonValidator.is_string(filename):
        return False

    # TODO: I cannot import from 'file.filename' because of
    # cyclic import issue
    _, extension = os.path.splitext(filename)

    if '.' in extension:
        extension = extension.replace('.', '')

    if extension is None:
        return False
    
    return file_extension_enum_class.is_valid(extension)

def get_temp_filename_for_file_extension(
    filename: Union[str, None] = None,
    file_extension: Union[FileExtension, TextFileExtension, AudioFileExtension, ImageFileExtension, VideoFileExtension, SubtitleFileExtension] = FileExtension
) -> str:
    """
    Get a temporary random filename with this file
    extension.

    The 'filename', if provided, will be forced to
    have this file extension but keeping the rest.
    If it is not provided, it will be a randomly
    generated string with this file extension.
    """
    if not PythonValidator.is_instance(file_extension, [FileExtension, TextFileExtension, AudioFileExtension, ImageFileExtension, VideoFileExtension, SubtitleFileExtension]):
        raise Exception('The "file_extension_enum" is not valid.')

    filename = Temp.create_filename(f'{Random.characters()}.noext') if filename is None else filename
    filename = filename.split('.')[0] if '.' in filename else filename
    filename += f'.{file_extension.value}'

    return filename


from yta_general_utils.programming.validator import PythonValidator
from typing import Union
from dataclasses import dataclass

import inspect


@dataclass
class MethodParameter:
    """
    The representation of a parameter found in a
    method signature, indicating if it was what 
    we consider an optional parameter or a 
    mandatory one based on its default value.
    """

    name: str
    """
    The name of the parameter.
    """
    type: any
    """
    The type of the parameter.
    """
    default_value: any
    """
    The default value of the parameter.
    """

    def __init__(
        self,
        name: str,
        type: any,
        default_value: any
    ):
        self.name = name
        # TODO: The type, that comes from '.annotation' has
        # to be reconsidered as it is quite special
        self.type = type
        self.default_value = default_value

    @property
    def is_default_value_empty(
        self
    ) -> bool:
        """
        Check if the default value is empty (no value
        associated) or  not.
        """
        return self.default_value is inspect._empty
    
    @property
    def is_optional(
        self
    ) -> bool:
        """
        Check if the parameter is optional (its default
        value is None).
        """
        return self.default_value is None
    
    @property
    def is_mandatory(
        self
    ) -> bool:
        """
        Check if the parameter is mandatory (its default
        value is not None).
        """
        return not self.is_optional
    
    @property
    def as_dict(
        self
    ) -> dict:
        """
        Get the parameter as a dict with its 'name' as
        the key and the instance as the value.

        'name' : self
        """
        return {
            self.name: self
        }
    
@dataclass
class MethodParameters:
    """
    List of parameters that exist in a function
    or method signature, to be able to handle
    them easy and detect which ones are mandatory
    or optional and more stuff.
    """

    parameters: list[MethodParameter]
    """
    The parameters found in the method signature.
    """

    def __init__(
        self,
        parameters: list[MethodParameter]
    ):
        if (
            not PythonValidator.is_list(parameters) or
            any(not PythonValidator.is_instance(item, MethodParameter) for item in parameters)
        ):
            raise Exception('The provided "parameters" parameter is not a list of MethodParameter instances or at least one of the elements is not a MethodParameter instance.')

        self.parameters = parameters
        
    @property
    def mandatory(
        self
    ) -> 'MethodParameters':
        """
        Get the parameters that are mandatory as a new
        instance of a MethodParameters class containing
        only those ones.

        This is a new instance that doesn't modify the
        original one.
        """
        return MethodParameters(self.mandatory_as_list)
    
    @property
    def optional(
        self
    ) -> 'MethodParameters':
        """
        Get the parameters that are optional as a new
        instance of a MethodParameters class containing
        only those ones.

        This is a new instance that doesn't modify the
        original one.
        """
        return MethodParameters(self.optional_as_list)

    @property
    def mandatory_as_list(
        self
    ) -> list[MethodParameter]:
        """
        Get the parameters that are mandatory as a list
        of MethodParameter instances.
        """
        return [
            parameter
            for parameter in self.parameters
            if parameter.is_mandatory
        ]
    
    @property
    def optional_as_list(
        self
    ) -> list[MethodParameter]:
        """
        Get the parameters that are optional as a list
        of MethodParameter instances.
        """
        return [
            parameter
            for parameter in self.parameters
            if parameter.is_optional
        ]
    
    @property
    def as_dict(
        self
    ) -> dict[str, MethodParameter]:
        """
        Get the parameters as a dictionary in which the
        parameter names are the keys and the values are
        MethodParameter instances.
        """
        if (
            not hasattr(self, '_as_dict') or
            self._as_dict is None
        ):
            self._as_dict = {
                parameter.name: parameter
                for parameter in self.parameters
            }

        return self._as_dict
    
    def get(
        self,
        name: str
    ) -> Union[MethodParameter, None]:
        """
        Get the parameter with the given 'name' if
        existing, or None if not.
        """
        return self.as_dict.get(name, None)
    
    def has(
        self,
        name: str
    ) -> bool:
        """
        Check if the parameter with the given 'name' 
        exist or not.
        """
        return name in self.as_dict
    
    def remove(
        self,
        name: str
    ) -> 'Parameters':
        """
        Remove the parameter with the provided 'name' if
        existing.
        """
        self.parameters = [
            parameter
            for parameter in self.parameters
            if parameter.name != name
        ]

        # We force it to be recalculated
        self._as_dict = None

        return self



from yta_general_utils.file.enums import FileTypeX, FileExtension
from yta_general_utils.programming.validator import PythonValidator
from typing import Union


class Output:
    """
    Class to handle the outputs we need for our app
    and force the desired extension if needed.
    """

    def get_filename(
        output_filename: Union[None, str],
        extension: Union[None, FileExtension, FileTypeX, str] = None
    ):
        """
        This method handles the provided 'output_filename' and
        the also given 'extension' to obtain the correct
        filename if needed. This has been created to simplify
        the way we handle the 'output_filename' param when
        working with our different libraries.

        Providing a FileType instance as 'extension' parameter
        will check if the provided 'output_filename' has an
        extension of that file type and preserve it, or use
        the default one if filename extension doesn't fit the
        desired type (or even the filename doesn't have any
        extension). Providing a FileExtension instance as
        'extension' parameter will force that extension for
        the generated output filename.
        """
        # 1. We don't need output and we don't pass output
        if (
            output_filename is None and
            extension is None
        ):
            return None
        
        # 3. We don't need output but we pass output
        if (
            output_filename is not None and
            extension is None
        ):
            # TODO: Maybe validate if extension in str (?)
            return output_filename
        
        # 2. We need output but we don't pass output file
        # 4. We need output and we pass output
        else:
            # Handle extension, as it is set
            if PythonValidator.is_string(extension):
                if FileExtension.is_valid_name_or_value(extension):
                    extension = FileExtension.to_enum(extension)
                elif FileTypeX.is_valid_name_or_value(extension):
                    extension = FileTypeX.to_enum(extension)
                else:
                    raise Exception('The provided "exception" is an invalid string.')
            elif not PythonValidator.is_instance(extension, [FileExtension, FileTypeX]):
                raise Exception('The provided "extension" is not a FileExtension nor a FileTypeX instance.')
            
            output_filename = (
                output_filename
                if PythonValidator.is_instance(extension, FileTypeX) and extension.is_filename_valid(output_filename)
                else extension.get_temp_filename(output_filename)
            )

            return output_filename


from yta_general_utils.web.tiktok.url_parser import TiktokUrlParser
from yta_general_utils.dataclasses import FileReturn
from yta_general_utils.downloader import Downloader
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.programming.output import Output
from yta_general_utils.programming.validator import PythonValidator
from typing import Union


DOWNLOAD_CDN_URL = 'https://tikcdn.io/ssstik/' # + video_id to download

# TODO: Make this work also with video_id (?)
def get_tiktok_video(
    url: str,
    output_filename: Union[str, None] = None
) -> FileReturn:
    """
    Obtains the Tiktok video from the provided 'url' if 
    valid and stores it locally if 'output_filename' is
    provided, or returns it if not.
    """
    if not PythonValidator.is_string(url):
        raise Exception('The provided "url" parameter is not a string.')

    tiktok_video_info = TiktokUrlParser.parse(url)

    download_url = f'{DOWNLOAD_CDN_URL}{tiktok_video_info.video_id}'

    return Downloader.download_video(
        download_url,
        Output.get_filename(output_filename, FileTypeX.VIDEO)
    )

    # TODO: I think this is being done by the line above
    # if output_filename:
    #     FileWriter.write_binary_file(video_content, output_filename)

    return video_content

# TODO: Make this work also with video_id (?)
def download_tiktok_video(
    url: str,
    output_filename: str
) -> FileReturn:
    """
    Obtains the Tiktok video from the provided 'url' if 
    valid and stores it locally as 'output_filename'.
    """
    if not PythonValidator.is_string(url):
        raise Exception('The provided "url" parameter is not a string.')
    
    if not output_filename:
        raise Exception('No "output_filename" provided to save the file.')
    
    return get_tiktok_video(url, output_filename)


from yta_general_utils.programming.validator import PythonValidator
from typing import Union

import numpy as np


class NumberValidator:
    """
    Class to simplify and encapsulate the functionality
    related to validate numeric values.
    """

    @staticmethod
    def is_number(
        element: Union[int, float, str, np.number],
        do_accept_string_number: bool = False
    ) -> bool:
        """
        Check if the provided 'element' is a numeric value. If
        'do_accept_string_number' is True, it will try to parse
        the 'element' as a float if a string is provided.
        """
        if not PythonValidator.is_instance(element, [int, float, str, np.number]):
            return False
        
        if PythonValidator.is_instance(element, str):
            if do_accept_string_number:
                try:
                    float(element)
                except:
                    return False
            else:
                return False
            
        return True
    
    @staticmethod
    def is_positive_number(
        element: Union[int, float, str, np.number],
        do_include_zero: bool = True
    ) -> bool:
        """
        This method checks if the provided 'element' is a numeric type,
        or tries to cast it as a float number if string provided, and
        returns True in the only case that the 'element' is actual a
        number by itself or as a string and it is 0 or above it. If 
        'do_include_zero' is set to False it won't be included.
        """        
        if not NumberValidator.is_number(element, False):
            return False
        
        element = float(element)

        return (
            element >= 0
            if do_include_zero else
            element > 0
        )
    
    @staticmethod
    def is_number_between(
        element: Union[int, float, str, np.number],
        lower_limit: Union[int, float, str, np.number],
        upper_limit: Union[int, float, str, np.number],
        do_include_lower_limit: bool = True,
        do_include_upper_limit: bool = True
    ) -> bool:
        """
        This methods returns True if the provided 'variable' is a valid number
        that is between the also provided 'lower_limit' and 'upper_limit'. It
        will return False in any other case.
        """
        if not NumberValidator.is_number(element, True):
            return False
        
        if not NumberValidator.is_number(lower_limit) or not NumberValidator.is_number(upper_limit):
            return False
        
        element = float(element)
        lower_limit = float(lower_limit)
        upper_limit = float(upper_limit)
        
        # TODO: Should we switch limits if unordered (?)
        # if upper_limit < lower_limit:
        #     raise Exception(f'The provided "upper_limit" parameter {str(upper_limit)} is lower than the "lower_limit" parameter {str(lower_limit)} provided.')

        if do_include_lower_limit and do_include_upper_limit:
            return lower_limit <= element <= upper_limit
        elif do_include_lower_limit:
            return lower_limit <= element < upper_limit
        elif do_include_upper_limit:
            return lower_limit < element <= upper_limit
        else:
            return lower_limit < element < upper_limit
        
    @staticmethod
    def is_int(
        element: int
    ) -> bool:
        """
        Return True if the provided 'element' is an int
        number.
        """
        return PythonValidator.is_instance(element, int)
    
    @staticmethod
    def is_float(
        element: float
    ) -> bool:
        """
        Return True if the provided 'element' is a float
        number.
        """
        return PythonValidator.is_instance(element, float)

    @staticmethod
    def is_even(
        element: float
    ) -> bool:
        """
        Return True if the provided 'element' is an even
        number. This method considers that the provided
        'element' is a valid number.
        """
        return element % 2 == 0
    
    @staticmethod
    def is_odd(
        element: float
    ) -> bool:
        """
        Return True if the provided 'element' is an odd
        number. This method considers that the provided
        'element' is a valid number.
        """
        return element % 2 != 0
    
